@@ -22,7 +22,9 @@
 #include "soup-marshal.h"
 #include "soup-message.h"
 #include "soup-message-private.h"
+#include "soup-message-queue.h"
 #include "soup-misc.h"
+#include "soup-misc-private.h"
 #include "soup-socket.h"
 #include "soup-ssl.h"
 #include "soup-uri.h"
@@ -35,10 +37,11 @@ typedef struct {
 	SoupURI     *proxy_uri;
 	gpointer     ssl_creds;
 	gboolean     ssl_strict;
+	gboolean     ssl_fallback;
 
 	GMainContext      *async_context;
 
-	SoupMessage *cur_req;
+	SoupMessageQueueItem *cur_item;
 	SoupConnectionState state;
 	time_t       unused_timeout;
 	guint        io_timeout, idle_timeout;
@@ -63,6 +66,7 @@ enum {
 	PROP_PROXY_URI,
 	PROP_SSL_CREDS,
 	PROP_SSL_STRICT,
+	PROP_SSL_FALLBACK,
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_IDLE_TIMEOUT,
@@ -78,7 +82,7 @@ static void get_property (GObject *object, guint prop_id,
 			  GValue *value, GParamSpec *pspec);
 
 static void stop_idle_timer (SoupConnectionPrivate *priv);
-static void clear_current_request (SoupConnection *conn);
+static void clear_current_item (SoupConnection *conn);
 
 /* Number of seconds after which we close a connection that hasn't yet
  * been used.
@@ -116,11 +120,17 @@ dispose (GObject *object)
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
 	stop_idle_timer (priv);
-	/* Make sure clear_current_request doesn't re-establish the timeout */
+	/* Make sure clear_current_item doesn't re-establish the timeout */
 	priv->idle_timeout = 0;
 
-	clear_current_request (conn);
-	soup_connection_disconnect (conn);
+	if (priv->cur_item) {
+		g_warning ("Disposing connection with cur_item set");
+		clear_current_item (conn);
+	}
+	if (priv->socket) {
+		g_warning ("Disposing connection while connected");
+		soup_connection_disconnect (conn);
+	}
 
 	G_OBJECT_CLASS (soup_connection_parent_class)->dispose (object);
 }
@@ -183,6 +193,13 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 				      "Whether certificate errors should be considered a connection error",
 				      TRUE,
 				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (
+		object_class, PROP_SSL_FALLBACK,
+		g_param_spec_boolean (SOUP_CONNECTION_SSL_FALLBACK,
+				      "SSLv3 fallback",
+				      "Use SSLv3 instead of TLS",
+				      FALSE,
+				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (
 		object_class, PROP_ASYNC_CONTEXT,
 		g_param_spec_pointer (SOUP_CONNECTION_ASYNC_CONTEXT,
@@ -258,6 +275,9 @@ set_property (GObject *object, guint prop_id,
 	case PROP_SSL_STRICT:
 		priv->ssl_strict = g_value_get_boolean (value);
 		break;
+	case PROP_SSL_FALLBACK:
+		priv->ssl_fallback = g_value_get_boolean (value);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		priv->async_context = g_value_get_pointer (value);
 		if (priv->async_context)
@@ -300,6 +320,9 @@ get_property (GObject *object, guint prop_id,
 	case PROP_SSL_STRICT:
 		g_value_set_boolean (value, priv->ssl_strict);
 		break;
+	case PROP_SSL_FALLBACK:
+		g_value_set_boolean (value, priv->ssl_fallback);
+		break;
 	case PROP_ASYNC_CONTEXT:
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
@@ -313,7 +336,10 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_enum (value, priv->state);
 		break;
 	case PROP_MESSAGE:
-		g_value_set_object (value, priv->cur_req);
+		if (priv->cur_item)
+			g_value_set_object (value, priv->cur_item->msg);
+		else
+			g_value_set_object (value, NULL);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -351,57 +377,53 @@ stop_idle_timer (SoupConnectionPrivate *priv)
 }
 
 static void
-set_current_request (SoupConnection *conn, SoupMessage *req)
+set_current_item (SoupConnection *conn, SoupMessageQueueItem *item)
 {
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	g_return_if_fail (priv->cur_req == NULL);
+	g_return_if_fail (priv->cur_item == NULL);
 
 	g_object_freeze_notify (G_OBJECT (conn));
 
 	stop_idle_timer (priv);
-	priv->unused_timeout = 0;
 
-	soup_message_set_io_status (req, SOUP_MESSAGE_IO_STATUS_RUNNING);
-	priv->cur_req = req;
+	item->state = SOUP_MESSAGE_RUNNING;
+	priv->cur_item = item;
 	g_object_notify (G_OBJECT (conn), "message");
 
 	if (priv->state == SOUP_CONNECTION_IDLE ||
-	    req->method != SOUP_METHOD_CONNECT)
+	    item->msg->method != SOUP_METHOD_CONNECT)
 		soup_connection_set_state (conn, SOUP_CONNECTION_IN_USE);
 
-	g_object_add_weak_pointer (G_OBJECT (req), (gpointer)&priv->cur_req);
-
 	g_object_thaw_notify (G_OBJECT (conn));
 }
 
 static void
-clear_current_request (SoupConnection *conn)
+clear_current_item (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
 	g_object_freeze_notify (G_OBJECT (conn));
 
-	if (priv->state == SOUP_CONNECTION_IN_USE) {
-		/* We don't use soup_connection_set_state here since
-		 * it may call clear_current_request()...
-		 */
-		priv->state = SOUP_CONNECTION_IDLE;
-		g_object_notify (G_OBJECT (conn), "state");
-	}
+	priv->unused_timeout = 0;
 	start_idle_timer (conn);
-	if (priv->cur_req) {
-		SoupMessage *cur_req = priv->cur_req;
 
-		g_object_remove_weak_pointer (G_OBJECT (priv->cur_req),
-					      (gpointer)&priv->cur_req);
-		priv->cur_req = NULL;
+	if (priv->cur_item) {
+		SoupMessageQueueItem *item;
+
+		item = priv->cur_item;
+		priv->cur_item = NULL;
 		g_object_notify (G_OBJECT (conn), "message");
 
-		if (!soup_message_is_keepalive (cur_req))
+		if (item->msg->method == SOUP_METHOD_CONNECT &&
+		    SOUP_STATUS_IS_SUCCESSFUL (item->msg->status_code)) {
+			/* We're now effectively no longer proxying */
+			soup_uri_free (priv->proxy_uri);
+			priv->proxy_uri = NULL;
+		}
+
+		if (!soup_message_is_keepalive (item->msg))
 			soup_connection_disconnect (conn);
-		else
-			soup_message_io_stop (cur_req);
 	}
 
 	g_object_thaw_notify (G_OBJECT (conn));
@@ -417,50 +439,61 @@ typedef struct {
 	SoupConnection *conn;
 	SoupConnectionCallback callback;
 	gpointer callback_data;
+	GCancellable *cancellable;
 } SoupConnectionAsyncConnectData;
 
+static void
+socket_connect_finished (SoupSocket *socket, guint status, gpointer user_data)
+{
+	SoupConnectionAsyncConnectData *data = user_data;
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
+
+	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
+		g_signal_connect (priv->socket, "disconnected",
+				  G_CALLBACK (socket_disconnected), data->conn);
+
+		soup_connection_set_state (data->conn, SOUP_CONNECTION_IN_USE);
+		priv->unused_timeout = time (NULL) + SOUP_CONNECTION_UNUSED_TIMEOUT;
+		start_idle_timer (data->conn);
+	} else if (status == SOUP_STATUS_TLS_FAILED) {
+		priv->ssl_fallback = TRUE;
+		status = SOUP_STATUS_TRY_AGAIN;
+	}
+
+	if (data->callback) {
+		if (priv->proxy_uri != NULL)
+			status = soup_status_proxify (status);
+		data->callback (data->conn, status, data->callback_data);
+	}
+	g_object_unref (data->conn);
+	if (data->cancellable)
+		g_object_unref (data->cancellable);
+	g_slice_free (SoupConnectionAsyncConnectData, data);
+}
+
 static void
 socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 {
 	SoupConnectionAsyncConnectData *data = user_data;
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
 
-	if (!SOUP_STATUS_IS_SUCCESSFUL (status))
-		goto done;
-
-	if (priv->ssl_creds && !priv->tunnel_addr) {
-		if (!soup_socket_start_ssl (sock, NULL)) {
-			status = SOUP_STATUS_SSL_FAILED;
-			goto done;
+	if (SOUP_STATUS_IS_SUCCESSFUL (status) &&
+	    priv->ssl_creds && !priv->tunnel_addr) {
+		if (soup_socket_start_ssl (sock, data->cancellable)) {
+			soup_socket_handshake_async (sock, data->cancellable,
+						     socket_connect_finished, data);
+			return;
 		}
+
+		status = SOUP_STATUS_SSL_FAILED;
 	}
 
-	g_signal_connect (priv->socket, "disconnected",
-			  G_CALLBACK (socket_disconnected), data->conn);
-
-	soup_connection_set_state (data->conn, SOUP_CONNECTION_IDLE);
-	priv->unused_timeout = time (NULL) + SOUP_CONNECTION_UNUSED_TIMEOUT;
-	start_idle_timer (data->conn);
-
- done:
-	if (data->callback) {
-		if (priv->proxy_uri != NULL)
-			status = soup_status_proxify (status);
-		data->callback (data->conn, status, data->callback_data);
-	}
-	g_slice_free (SoupConnectionAsyncConnectData, data);
+	socket_connect_finished (sock, status, data);
 }
 
-/**
- * soup_connection_connect_async:
- * @conn: the connection
- * @callback: callback to call when the connection succeeds or fails
- * @user_data: data for @callback
- *
- * Asynchronously connects @conn.
- **/
 void
 soup_connection_connect_async (SoupConnection *conn,
+			       GCancellable *cancellable,
 			       SoupConnectionCallback callback,
 			       gpointer user_data)
 {
@@ -474,31 +507,26 @@ soup_connection_connect_async (SoupConnection *conn,
 	soup_connection_set_state (conn, SOUP_CONNECTION_CONNECTING);
 
 	data = g_slice_new (SoupConnectionAsyncConnectData);
-	data->conn = conn;
+	data->conn = g_object_ref (conn);
 	data->callback = callback;
 	data->callback_data = user_data;
+	data->cancellable = cancellable ? g_object_ref (cancellable) : NULL;
 
 	priv->socket =
 		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, priv->remote_addr,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
 				 SOUP_SOCKET_SSL_STRICT, priv->ssl_strict,
+				 SOUP_SOCKET_SSL_FALLBACK, priv->ssl_fallback,
 				 SOUP_SOCKET_ASYNC_CONTEXT, priv->async_context,
 				 SOUP_SOCKET_TIMEOUT, priv->io_timeout,
+				 "clean-dispose", TRUE,
 				 NULL);
-	soup_socket_connect_async (priv->socket, NULL,
+	soup_socket_connect_async (priv->socket, cancellable,
 				   socket_connect_result, data);
 }
 
-/**
- * soup_connection_connect_sync:
- * @conn: the connection
- *
- * Synchronously connects @conn.
- *
- * Return value: the soup status
- **/
 guint
-soup_connection_connect_sync (SoupConnection *conn)
+soup_connection_connect_sync (SoupConnection *conn, GCancellable *cancellable)
 {
 	SoupConnectionPrivate *priv;
 	guint status;
@@ -513,32 +541,40 @@ soup_connection_connect_sync (SoupConnection *conn)
 		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, priv->remote_addr,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->ssl_creds,
 				 SOUP_SOCKET_SSL_STRICT, priv->ssl_strict,
+				 SOUP_SOCKET_SSL_FALLBACK, priv->ssl_fallback,
 				 SOUP_SOCKET_FLAG_NONBLOCKING, FALSE,
 				 SOUP_SOCKET_TIMEOUT, priv->io_timeout,
+				 "clean-dispose", TRUE,
 				 NULL);
 
-	status = soup_socket_connect_sync (priv->socket, NULL);
+	status = soup_socket_connect_sync (priv->socket, cancellable);
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 		goto fail;
 		
-	g_signal_connect (priv->socket, "disconnected",
-			  G_CALLBACK (socket_disconnected), conn);
-
 	if (priv->ssl_creds && !priv->tunnel_addr) {
-		if (!soup_socket_start_ssl (priv->socket, NULL)) {
+		if (!soup_socket_start_ssl (priv->socket, cancellable))
 			status = SOUP_STATUS_SSL_FAILED;
-			goto fail;
+		else {
+			status = soup_socket_handshake_sync (priv->socket, cancellable);
+			if (status == SOUP_STATUS_TLS_FAILED) {
+				priv->ssl_fallback = TRUE;
+				status = SOUP_STATUS_TRY_AGAIN;
+			}
 		}
 	}
 
 	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
-		soup_connection_set_state (conn, SOUP_CONNECTION_IDLE);
+		g_signal_connect (priv->socket, "disconnected",
+				  G_CALLBACK (socket_disconnected), conn);
+
+		soup_connection_set_state (conn, SOUP_CONNECTION_IN_USE);
 		priv->unused_timeout = time (NULL) + SOUP_CONNECTION_UNUSED_TIMEOUT;
 		start_idle_timer (conn);
 	} else {
 	fail:
 		if (priv->socket) {
+			soup_socket_disconnect (priv->socket);
 			g_object_unref (priv->socket);
 			priv->socket = NULL;
 		}
@@ -560,11 +596,13 @@ soup_connection_get_tunnel_addr (SoupConnection *conn)
 	return priv->tunnel_addr;
 }
 
-gboolean
-soup_connection_start_ssl (SoupConnection *conn)
+guint
+soup_connection_start_ssl_sync (SoupConnection *conn,
+				GCancellable   *cancellable)
 {
 	SoupConnectionPrivate *priv;
 	const char *server_name;
+	guint status;
 
 	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
@@ -572,7 +610,74 @@ soup_connection_start_ssl (SoupConnection *conn)
 	server_name = soup_address_get_name (priv->tunnel_addr ?
 					     priv->tunnel_addr :
 					     priv->remote_addr);
-	return soup_socket_start_proxy_ssl (priv->socket, server_name, NULL);
+	if (!soup_socket_start_proxy_ssl (priv->socket, server_name,
+					  cancellable))
+		return SOUP_STATUS_SSL_FAILED;
+
+	status = soup_socket_handshake_sync (priv->socket, cancellable);
+	if (status == SOUP_STATUS_TLS_FAILED) {
+		priv->ssl_fallback = TRUE;
+		status = SOUP_STATUS_TRY_AGAIN;
+	}
+
+	return status;
+}
+
+static void
+start_ssl_completed (SoupSocket *socket, guint status, gpointer user_data)
+{
+	SoupConnectionAsyncConnectData *data = user_data;
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
+
+	if (status == SOUP_STATUS_TLS_FAILED) {
+		priv->ssl_fallback = TRUE;
+		status = SOUP_STATUS_TRY_AGAIN;
+	}
+
+	data->callback (data->conn, status, data->callback_data);
+	g_object_unref (data->conn);
+	g_slice_free (SoupConnectionAsyncConnectData, data);
+}
+
+static gboolean
+idle_start_ssl_completed (gpointer user_data)
+{
+	SoupConnectionAsyncConnectData *data = user_data;
+
+	start_ssl_completed (NULL, SOUP_STATUS_SSL_FAILED, data);
+	return FALSE;
+}
+
+void
+soup_connection_start_ssl_async (SoupConnection   *conn,
+				 GCancellable     *cancellable,
+				 SoupConnectionCallback callback,
+				 gpointer          user_data)
+{
+	SoupConnectionPrivate *priv;
+	const char *server_name;
+	SoupConnectionAsyncConnectData *data;
+
+	g_return_if_fail (SOUP_IS_CONNECTION (conn));
+	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
+
+	data = g_slice_new (SoupConnectionAsyncConnectData);
+	data->conn = g_object_ref (conn);
+	data->callback = callback;
+	data->callback_data = user_data;
+
+	server_name = soup_address_get_name (priv->tunnel_addr ?
+					     priv->tunnel_addr :
+					     priv->remote_addr);
+	if (!soup_socket_start_proxy_ssl (priv->socket, server_name,
+					  cancellable)) {
+		soup_add_completion (priv->async_context,
+				     idle_start_ssl_completed, data);
+		return;
+	}
+
+	soup_socket_handshake_async (priv->socket, cancellable,
+				     start_ssl_completed, data);
 }
 
 /**
@@ -586,26 +691,25 @@ void
 soup_connection_disconnect (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv;
+	SoupConnectionState old_state;
 
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	if (!priv->socket)
-		return;
+	old_state = priv->state;
+	if (old_state != SOUP_CONNECTION_DISCONNECTED)
+		soup_connection_set_state (conn, SOUP_CONNECTION_DISCONNECTED);
 
-	g_signal_handlers_disconnect_by_func (priv->socket,
-					      socket_disconnected, conn);
-	soup_socket_disconnect (priv->socket);
-	g_object_unref (priv->socket);
-	priv->socket = NULL;
+	if (priv->socket) {
+		g_signal_handlers_disconnect_by_func (priv->socket,
+						      socket_disconnected, conn);
+		soup_socket_disconnect (priv->socket);
+		g_object_unref (priv->socket);
+		priv->socket = NULL;
+	}
 
-	/* Don't emit "disconnected" if we aren't yet connected */
-	if (priv->state < SOUP_CONNECTION_IDLE)
-		return;
-
-	soup_connection_set_state (conn, SOUP_CONNECTION_DISCONNECTED);
-	/* NB: this might cause conn to be destroyed. */
-	g_signal_emit (conn, signals[DISCONNECTED], 0);
+	if (old_state != SOUP_CONNECTION_DISCONNECTED)
+		g_signal_emit (conn, signals[DISCONNECTED], 0);
 }
 
 SoupSocket *
@@ -624,6 +728,14 @@ soup_connection_get_proxy_uri (SoupConnection *conn)
 	return SOUP_CONNECTION_GET_PRIVATE (conn)->proxy_uri;
 }
 
+gboolean
+soup_connection_is_via_proxy (SoupConnection *conn)
+{
+	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
+
+	return SOUP_CONNECTION_GET_PRIVATE (conn)->proxy_uri != NULL;
+}
+
 SoupConnectionState
 soup_connection_get_state (SoupConnection *conn)
 {
@@ -661,36 +773,48 @@ soup_connection_set_state (SoupConnection *conn, SoupConnectionState state)
 	g_return_if_fail (state >= SOUP_CONNECTION_NEW &&
 			  state <= SOUP_CONNECTION_DISCONNECTED);
 
+	g_object_freeze_notify (G_OBJECT (conn));
+
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 	old_state = priv->state;
 	priv->state = state;
-	if (state == SOUP_CONNECTION_IDLE &&
+	if ((state == SOUP_CONNECTION_IDLE ||
+	     state == SOUP_CONNECTION_DISCONNECTED) &&
 	    old_state == SOUP_CONNECTION_IN_USE)
-		clear_current_request (conn);
+		clear_current_item (conn);
 
 	g_object_notify (G_OBJECT (conn), "state");
+	g_object_thaw_notify (G_OBJECT (conn));
+}
+
+gboolean
+soup_connection_get_ever_used (SoupConnection *conn)
+{
+	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
+
+	return SOUP_CONNECTION_GET_PRIVATE (conn)->unused_timeout == 0;
+}
+
+gboolean
+soup_connection_get_ssl_fallback (SoupConnection *conn)
+{
+	return SOUP_CONNECTION_GET_PRIVATE (conn)->ssl_fallback;
 }
 
-/**
- * soup_connection_send_request:
- * @conn: a #SoupConnection
- * @req: a #SoupMessage
- *
- * Sends @req on @conn. This is a low-level function, intended for use
- * by #SoupSession.
- **/
 void
-soup_connection_send_request (SoupConnection *conn, SoupMessage *req)
+soup_connection_send_request (SoupConnection          *conn,
+			      SoupMessageQueueItem    *item,
+			      SoupMessageCompletionFn  completion_cb,
+			      gpointer                 user_data)
 {
 	SoupConnectionPrivate *priv;
 
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
-	g_return_if_fail (SOUP_IS_MESSAGE (req));
+	g_return_if_fail (item != NULL);
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 	g_return_if_fail (priv->state != SOUP_CONNECTION_NEW && priv->state != SOUP_CONNECTION_DISCONNECTED);
 
-	if (req != priv->cur_req)
-		set_current_request (conn, req);
-	soup_message_send_request (req, priv->socket, conn,
-				   priv->proxy_uri != NULL);
+	if (item != priv->cur_item)
+		set_current_item (conn, item);
+	soup_message_send_request (item, completion_cb, user_data);
 }
